#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division

import sys
sys.path.insert(0, '..')

import numpy as np
from mpi4py import MPI
from matplotlib import use

from pulp.utils.parallel import pprint

from pulp.utils.datalog import dlog, StoreToH5, TextPrinter
from pulp.visualize.gui import GUI, RFViewer, YTPlotter

from pulp.em import EM
from pulp.em.annealing import LinearAnnealing
from pulp.em.camodels.linca_et import LinCA_ET

from random import sample


# Parameters
#============================================================================

# Number of data points:
N = 1000 
# Squareroot of dimensionality and dimensionality
D2 = 5
D = D2**2
# Number of hidden causes
H = 2*D2
# Value of bars in basis functions:
bar_val = 10.
# Percentage of negative bars in ground truth basis functions
neg_bars = 0.5

# Approximation parameters
Hprime = 4
gamma = 3

# Annealing schedule
Tsteps = 50
Tstart = 1.0
Tend = 1.0

#============================================================================
# Main

if __name__ == "__main__":
    comm = MPI.COMM_WORLD

    pprint("="*78)
    pprint(" Running %d parallel processes" % comm.size) 
    pprint("="*78)

    #Configure DataLogger
    #use('GTKAgg')
    dlog.start_gui(GUI)

    dlog.set_handler('L', YTPlotter)
    dlog.set_handler(('T', 'L', 'pi', 'sigma', 'N'), TextPrinter)
    dlog.set_handler(('W', 'pi', 'sigma', 'N', 'L'), StoreToH5, 'output/result.h5')
    dlog.set_handler('W', RFViewer, rf_shape=(D2, D2), symmetric=1, global_maximum=1, colormap='jet')
    #dlog.set_handler('y', RFViewer, rf_shape=(D2, D2))

    # Choose annealing schedule
    anneal = LinearAnnealing(Tsteps)
    anneal['T']           = [(10, Tstart) , (-10, Tend)]
    anneal['Ncut_factor'] = [(2./3, 0.), (-10, 1.)]
    anneal['W_noise']     = [(-10, 0.01), (-1, 0.)]

    # Prepare ground-truth GFs (bars)
    W_gt_tmp = np.zeros( (H, D2, D2) )
    for i in xrange(D2):
        W_gt_tmp[   i, i, :] = bar_val
        W_gt_tmp[D2+i, :, i] = bar_val
    if neg_bars > 0.0:
        W_gt_tmp[sample(range(H), np.int(H*neg_bars))] *= -1
    W_gt = comm.bcast(W_gt_tmp.reshape( (H, D) ))
    
    # Prepare model...
    model = LinCA_ET(D, H, Hprime, gamma)
    # Ground truh parameters
    gt_params = {
        'W'     : W_gt,
        'pi'    : 2./H,
        'sigma' : 2.00
    }
    
    # Generate trainig data
    my_N = N // comm.size
    my_data = model.generate_data(gt_params, my_N)
    # Append first 25 pictures to data logger for plotting
    dlog.append('y',my_data['y'][0:25,:])
    
    # Initialize model parameters (to be learned)
    params = {
        #'W'     : W_gt,
        'W'     : np.random.normal(size=W_gt.shape),
        'pi'    : 2/H, 
        'sigma' : 5.00
    }
    # Broadcast parameters to all CPUs
    params = comm.bcast(params)

   # Create and start EM annealing
    em = EM(model=model, anneal=anneal)
    em.data = my_data
    em.lparams = params
    em.run()
    
    dlog.close()
